import os
import hashlib
import json
from pathlib import Path
from .exceptions import PFXValidationError

def calculate_checksum(data: str) -> str:
    """Calculate MD5 checksum for data"""
    return hashlib.md5(data.encode('utf-8')).hexdigest()

def validate_pfx(file_path: str) -> bool:
    """Validate a PFX file structure"""
    try:
        with open(file_path, 'r') as f:
            content = f.read()

        # Basic validation checks
        if "# PandaFlex" not in content:
            raise PFXValidationError("Missing PandaFlex header")

        if "[Schema]" not in content or "[Data]" not in content:
            raise PFXValidationError("Missing required sections")

        return True
    except Exception as e:
        raise PFXValidationError(f"Validation failed: {str(e)}")

def generate_example(output_dir: str = ".", filename: str = "example.pfx"):
    """Generate an example PFX file"""
    example_content = """# PandaFlex v1.0
# Metadata: {"created": "2023-11-15", "description": "Example PandaFlex file"}

[Schema]
id: int32
name: str
value: float64
is_valid: bool
categories: list[str]
attributes: dict

[Data]
# Sample data
1, "Product A", 19.99, true, ["electronics", "gadgets"], {"weight": 0.5, "color": "black"}
2, "Product B", 29.99, false, ["home", "furniture"], {"weight": 5.2, "color": "brown"}
3, "Product C", 9.99, true, ["office", "supplies"], {"weight": 0.1, "color": "white"}

[Footer]
row_count: 3
compression: none
checksum: "example_checksum"
"""

    output_path = os.path.join(output_dir, filename)
    with open(output_path, 'w') as f:
        f.write(example_content)

    return output_path


class PFXError(Exception):
    """Base class for all PandaFlex exceptions"""
    pass

class PFXParseError(PFXError):
    """Raised when there's an error parsing a PFX file"""
    pass

class PFXValidationError(PFXError):
    """Raised when a PFX file fails validation"""
    pass

class PFXSchemaError(PFXError):
    """Raised when there's a schema mismatch"""
    pass


import json
import pandas as pd
from typing import Dict, Any, Optional
from .utils import calculate_checksum

def infer_schema(df: pd.DataFrame) -> Dict[str, str]:
    """Infer PFX schema from pandas DataFrame"""
    schema = {}

    for column, dtype in df.dtypes.items():
        if pd.api.types.is_integer_dtype(dtype):
            if dtype == 'int32':
                schema[column] = 'int32'
            else:
                schema[column] = 'int64'
        elif pd.api.types.is_float_dtype(dtype):
            schema[column] = 'float64'
        elif pd.api.types.is_bool_dtype(dtype):
            schema[column] = 'bool'
        elif pd.api.types.is_string_dtype(dtype):
            schema[column] = 'str'
        elif pd.api.types.is_list_like(dtype):
            # Try to infer list type from first non-null value
            non_null = df[column].dropna()
            if len(non_null) > 0:
                first = non_null.iloc[0]
                if isinstance(first, list) and len(first) > 0:
                    sample_type = type(first[0]).__name__
                    if sample_type == 'str':
                        schema[column] = 'list[str]'
                    elif sample_type == 'int':
                        schema[column] = 'list[int32]'
                    elif sample_type == 'float':
                        schema[column] = 'list[float64]'
                    else:
                        schema[column] = f'list[{sample_type}]'
                else:
                    schema[column] = 'list[str]'  # default
            else:
                schema[column] = 'list[str]'  # default
        elif pd.api.types.is_dict_like(dtype):
            schema[column] = 'dict'
        else:
            schema[column] = 'str'  # fallback

    return schema

def format_value(value: Any) -> str:
    """Format a value for PFX output"""
    if value is None:
        return 'null'
    elif isinstance(value, bool):
        return 'true' if value else 'false'
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, str):
        return f'"{value.replace("\\", "\\\\").replace("\"", "\\\"")}"'
    elif isinstance(value, list):
        return json.dumps(value)
    elif isinstance(value, dict):
        return json.dumps(value)
    else:
        return str(value)

def to_pfx(df: pd.DataFrame, file_path: str, metadata: Optional[Dict] = None,
           compression: str = 'none', description: str = ""):
    """
    Write a pandas DataFrame to a PandaFlex file

    Args:
        df: pandas DataFrame to save
        file_path: Path to save the PFX file
        metadata: Optional metadata dictionary
        compression: Compression type (currently only 'none' supported)
        description: Description of the dataset
    """
    # Infer schema
    schema = infer_schema(df)

    # Prepare header
    header = [
        "# PandaFlex v1.0",
        f'# Metadata: {json.dumps({"description": description} if description else {})}'
    ]

    # Prepare schema section
    schema_lines = ["[Schema]"] + [f"{col}: {dtype}" for col, dtype in schema.items()]

    # Prepare data section
    data_lines = ["[Data]"]
    for _, row in df.iterrows():
        row_values = [format_value(row[col]) for col in df.columns]
        data_lines.append(", ".join(row_values))

    # Prepare footer
    footer = [
        "[Footer]",
        f"row_count: {len(df)}",
        f"compression: {compression}",
        f"checksum: \"{calculate_checksum('\n'.join(data_lines))}\""
    ]

    # Combine all sections
    content = "\n".join(header + [""] + schema_lines + [""] + data_lines + [""] + footer)

    # Write to file
    with open(file_path, 'w') as f:
        f.write(content)

    # Add metadata if provided
    if metadata is not None:
        with open(file_path, 'a') as f:
            f.write(f"\n\n[Metadata]\n{json.dumps(metadata, indent=2)}")


import re
import json
import pandas as pd
from typing import Dict, Any, Union
from .exceptions import PFXParseError, PFXSchemaError
from .utils import validate_pfx

def parse_schema(schema_section: str) -> Dict[str, str]:
    """Parse the schema section into a dictionary"""
    schema = {}
    for line in schema_section.split('\n'):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        try:
            field_name, field_type = line.split(':')
            schema[field_name.strip()] = field_type.strip()
        except ValueError:
            raise PFXParseError(f"Invalid schema line: {line}")
    return schema

def parse_data(data_section: str, schema: Dict[str, str]) -> Dict[str, list]:
    """Parse the data section into a dictionary of columns"""
    data = {field: [] for field in schema}
    lines = [line.strip() for line in data_section.split('\n') if line.strip() and not line.startswith('#')]

    for line in lines:
        # Remove comments
        line = re.sub(r'#.*$', '', line).strip()
        if not line:
            continue

        # Parse values
        values = []
        current = ""
        in_quotes = False
        escape = False

        for char in line:
            if escape:
                current += char
                escape = False
            elif char == '\\':
                escape = True
            elif char == '"':
                in_quotes = not in_quotes
                current += char
            elif char == ',' and not in_quotes:
                values.append(current.strip())
                current = ""
            else:
                current += char

        if current:
            values.append(current.strip())

        if len(values) != len(schema):
            raise PFXParseError(f"Row has {len(values)} values but schema expects {len(schema)}")

        # Convert values based on schema
        for (field, field_type), value in zip(schema.items(), values):
            try:
                if field_type.startswith('list['):
                    # Parse list
                    inner_type = field_type[5:-1]
                    if value.startswith('[') and value.endswith(']'):
                        items = json.loads(value)
                        converted = [convert_value(inner_type, item) for item in items]
                    else:
                        raise PFXParseError(f"Invalid list format for {field}: {value}")
                elif field_type.startswith('dict'):
                    # Parse dict
                    if value.startswith('{') and value.endswith('}'):
                        converted = json.loads(value)
                    else:
                        raise PFXParseError(f"Invalid dict format for {field}: {value}")
                else:
                    # Parse primitive type
                    converted = convert_value(field_type, value)
                data[field].append(converted)
            except Exception as e:
                raise PFXParseError(f"Error converting {field} ({field_type}) with value {value}: {str(e)}")

    return data

def convert_value(field_type: str, value: str) -> Any:
    """Convert a string value to the appropriate Python type"""
    if value.lower() == 'true':
        return True
    elif value.lower() == 'false':
        return False
    elif value.lower() == 'null':
        return None

    if field_type == 'str':
        return value.strip('"')
    elif field_type == 'int32':
        return int(value)
    elif field_type == 'int64':
        return int(value)
    elif field_type == 'float32' or field_type == 'float64':
        return float(value)
    elif field_type == 'bool':
        return value.lower() == 'true'
    else:
        raise PFXParseError(f"Unsupported type: {field_type}")

def read_pfx(file_path: str) -> pd.DataFrame:
    """
    Read a PandaFlex file into a pandas DataFrame

    Args:
        file_path: Path to the PFX file

    Returns:
        pandas.DataFrame containing the data
    """
    # Validate the file first
    if not validate_pfx(file_path):
        raise PFXParseError("Invalid PFX file structure")

    with open(file_path, 'r') as f:
        content = f.read()

    # Split into sections
    sections = {}
    current_section = None

    for line in content.split('\n'):
        line = line.strip()
        if line.startswith('#'):
            continue
        elif line.startswith('[') and line.endswith(']'):
            current_section = line[1:-1]
            sections[current_section] = []
        elif current_section:
            sections[current_section].append(line)

    if 'Schema' not in sections or 'Data' not in sections:
        raise PFXParseError("Missing required sections in PFX file")

    # Parse schema
    schema = parse_schema('\n'.join(sections['Schema']))

    # Parse data
    data = parse_data('\n'.join(sections['Data']), schema)

    # Create DataFrame
    df = pd.DataFrame(data)

    # Parse metadata if exists
    if 'Metadata' in sections:
        metadata_str = '\n'.join(sections['Metadata'])
        try:
            metadata = json.loads(metadata_str)
            df.attrs['pandaflex_metadata'] = metadata
        except json.JSONDecodeError:
            pass

    return df


from setuptools import setup, find_packages
from pathlib import Path

# Read the contents of README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name="pandaflex",
    version="1.0.0",
    description="PandaFlex (PFX) - A new efficient data format for pandas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Louati Mahdi",
    author_email="louatimahdi390@gmail.com",
    url="https://github.com/mahdi123-tech",
    packages=find_packages(),
    install_requires=[
        "pandas>=1.0.0",
        "numpy>=1.18.0",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    python_requires=">=3.7",
    entry_points={
        'console_scripts': [
            'pandaflex=pandaflex.cli:main',
        ],
    },
    include_package_data=True,
    package_data={
        'pandaflex': ['examples/*.pfx'],
    },
)


from .version import __version__
from .reader import read_pfx
from .writer import to_pfx
from .utils import generate_example, validate_pfx
import os
from .utils import generate_example

# Generate example file on first import
if not os.path.exists('example.pfx'):
    generate_example()
    print("Generated example.pfx in your current directory. Check README.md for usage instructions.")

__all__ = ['read_pfx', 'to_pfx', 'generate_example', 'validate_pfx', '__version__']